"""

"""

import inspect


def init_args(cls):
    """ Return the __init__ args (minus 'self') for @cls

        Args:
            cls: class or instance
        Returns:
            list of str, the __init__ arguments minus 'self'
        Raises:
            ValueError if __init__ does not have a 'self' argument
    """
    argspec = inspect.getargspec(cls.__init__)
    args = argspec.args
    # raises ValueError if not present.  Not having a self argument
    # in __init__ is almost certainly a user error, and if not
    # there is a special place in hell for people who call the first
    # argument anything but 'self'
    args.remove('self')
    return args




"""

"""

from .init_args import init_args
from .key_swap import key_swap


__all__ = [
    'ExtraKeysError',
    'marshal_dict',
    'unmarshal_dict',
]


class ExtraKeysError(Exception):
    """ Raised when extra object keys are present """


def marshal_dict(
    obj,
    types,
):
    """ Recursively marshal a Python object to a JSON-compatible dict
        that can be passed to json.{dump,dumps}, a web client,
        or a web server, etc...

        Args:
            obj:   object, It's members can be nested Python
                   objects which will be converted to dictionaries
            types: tuple-of-types, The primitive types that can be
                   serialized
        Returns:
            dict
    """
    if getattr(obj, '_marshal_only_init_args', False):
        args = init_args(obj)
        excl = [x for x in obj.__dict__ if x not in args]
    else:
        excl = getattr(obj, '_marshal_exclude', [])

    if getattr(obj, '_marshal_exclude_none', False):
        excl.extend(k for k, v in obj.__dict__.items() if v is None)
    else:
        none_keys = getattr(obj, '_marshal_exclude_none_keys', [])
        if none_keys:
            excl.extend(x for x in none_keys if obj.__dict__.get(x) is None)

    return {
        k: v if isinstance(v, types) else marshal_dict(v, types)
        for k, v in obj.__dict__.items()
        if k not in excl
    }


def unmarshal_dict(
    obj,
    cls,
    allow_extra_keys=True,
):
    """ Unmarshal @obj into @cls

        Args:
            obj:              dict, The dict to unmarshal into @cls
            cls:              type, The class to unmarshal into
            allow_extra_keys: bool, False to raise an exception when extra
                              keys are present, True to ignore
        Returns:
            instance of @cls
        Raises:
            ExtraKeysError: If allow_extra_keys == False, and extra keys
                            are present in @obj and not in @cls.__init__
            ValueError:     If @cls.__init__ does not contain a self argument
    """
    args = init_args(cls)
    obj = key_swap(obj, cls, False)
    kwargs = {k: v for k, v in obj.items() if k in args}

    # If either is set to False, do not allow extra keys
    # to be present in obj but not in cls.__init__
    allow_extra_keys = (
        getattr(cls, '_unmarshal_allow_extra_keys', True)
        and
        allow_extra_keys
    )

    if not allow_extra_keys and len(obj) > len(kwargs):
        diff = {k: v for k, v in obj.items() if k not in args}
        msg = "Extra keys present, but allow_extra_keys=={}: {}".format(
            allow_extra_keys,
            diff,
        )
        raise ExtraKeysError(msg)

    return cls(**kwargs)




"""

"""

# Will raise an import error if the user hasn't installed 'bson'
import bson
import datetime

from .json import marshal_json
from .util.marshal import *
from .util.type import *


__all__ = [
    'bson',
    'ExtraKeysError',
    'marshal_bson',
    'MongoDocument',
    'type_assert',
    'type_assert_dict',
    'type_assert_iter',
    'unmarshal_bson',
]


BSON_TYPES = (
    bool,
    bson.ObjectId,
    datetime.datetime,
    dict,
    float,
    int,
    list,
    str,
    type(None),
)


class MongoDocument:
    """ Abstract class to facilitate inserting into MongoDB.
        Inherit this for classes that represent BSON documents.

        Assumes that you assigned the ObjectId in your
        class like:

            def __init__(
                self,
                ...,
                _id=None,
                ...,
            ):

            self._id = type_assert(
                _id,
                bson.ObjectId,
                allow_none=True,
            )
    """
    # If you need to override this, just add '_id' to the new list
    _marshal_exclude_none_keys = ['_id']

    def json(
        self,
        include_id=False,
    ):
        """ Helper method to convert to MongoDB documents to JSON

            The assumption is that you want to return results
            from a REST API that do not include the _id.

            If you have BSON types that are not JSON-compatible
            like datetime, override this method to cast them
            to JSON-compatible types

            Args:
                include_id: bool, True to cast _id to a str,
                            False to omit from the result
            Returns:
                dict
        """
        _id = self._id
        if include_id:
            self._id = str(self._id)
        else:
            self._id = None
        j = marshal_json(self)
        self._id = _id
        return j


def marshal_bson(
    obj,
    types=BSON_TYPES,
):
    """ Recursively marshal a Python object to a BSON-compatible dict
        that can be passed to PyMongo, Motor, etc...

        Args:
            obj:   object, It's members can be nested Python
                   objects which will be converted to dictionaries
            types: tuple-of-types, The BSON primitive types, typically
                   you would not change this
        Returns:
            dict
    """
    return marshal_dict(obj, types)


def unmarshal_bson(
    obj,
    cls,
    allow_extra_keys=True,
):
    """ Unmarshal @obj into @cls

        Args:
            obj:              dict, A BSON object
            cls:              type, The class to unmarshal into
            allow_extra_keys: bool, False to raise an exception when extra
                              keys are present, True to ignore
        Returns:
            instance of @cls
        Raises:
            ExtraKeysError: If allow_extra_keys == False, and extra keys
                            are present in @obj and not in @cls.__init__
            ValueError:     If @cls.__init__ does not contain a self argument
    """
    return unmarshal_dict(
        obj,
        cls,
        allow_extra_keys,
    )



"""

"""

from .marshal import unmarshal_dict
from .key_swap import key_swap


__all__ = [
    'type_assert',
    'type_assert_dict',
    'type_assert_iter',
]


def _check(
    obj,
    cls,
    allow_none=False,
):
    if (
        allow_none
        and
        obj is None
    ):
        return obj

    if not isinstance(obj, cls):
        if isinstance(obj, dict):
            obj = key_swap(obj, cls, True)
            return unmarshal_dict(obj, cls)
        msg = '{0} is not an instance of {1}, is {2}'.format(
            obj,
            cls,
            type(obj),
        )
        raise TypeError(msg)
    return obj


def type_assert(
    obj,
    cls,
    allow_none=False,
):
    """ Wrapper function for isinstance.

        Note that you cannot use this to unmarshal JSON if @cls
        is a tuple of types, it can only be a single type.

        Will also unmarshal JSON objects to Python objects if @obj
        is an instance of dict and @cls is not dict

        Args:
            obj: object instance, The object to type assert
            cls: type, The class type to assert
        Returns:
            @obj
        Raises:
            TypeError: if @obj is not an instance of @cls
    """
    return _check(
        obj,
        cls,
        allow_none,
    )


def type_assert_iter(
    iterable,
    cls,
):
    """ Checks that every object in @iterable is an instance of @cls

        Will also unmarshal JSON objects to Python objects if items in
        @iterable are an instance of dict

        Args:
            iterable:   Any iterable to check.  Note that it would not
                        make sense to pass a generator to this function
            cls:        type, The class type to assert
        Returns:
            @iterable, note that @iterable will be recreated, which
            may be a performance concern if @iterable has many items
        Raises:
            TypeError: if @obj is not an instance of @cls
    """
    t = type(iterable)
    return t(
        _check(obj, cls) for obj in iterable
    )


def type_assert_dict(
    d,
    kcls=None,
    vcls=None,
    allow_none=False,
):
    """ Checks that every key/value in @d is an instance of @kcls: @vcls

        Will also unmarshal JSON objects to Python objects if
        the value is an instance of dict and @vcls is a class type

        Args:
            d:           The dict to type assert
            @kcls:       The class to type assert for keys.
                         NOTE: JSON only allows str keys
            @vcls:       The class to type assert for values
            @allow_none: Allow a None value for the values.
                         This would not make sense for the keys.
        Returns:
            @d, note that @d will be recreated, which
            may be a performance concern if @d has many items
        Raises:
            TypeError: if a key is not an instance of @kcls or
                       a value is not an instance of @vcls
    """
    t = type(d)
    return t(
        (
            _check(k, kcls) if kcls else k,
            _check(v, vcls, allow_none) if vcls else v,
        )
        for k, v in d.items()
    )

